from typing import NamedTuple, Callable

import pytest
from sigdepsem import SigDepMeta, signature_dependent_semantics


class ClassSetup(NamedTuple):
    decorator: Callable[[type], type]
    meta: type


@pytest.fixture(params=[
    ClassSetup(signature_dependent_semantics, type),
    ClassSetup(lambda cls: cls, SigDepMeta),
])
def class_setup(request):
    return request.param


@pytest.fixture
def decorator(class_setup: ClassSetup):
    return class_setup.decorator


@pytest.fixture
def meta(class_setup: ClassSetup):
    return class_setup.meta



import pytest


@pytest.fixture
def C_class(decorator, meta):
    @decorator
    class C(metaclass=meta):
        def __init__(self, name):
            self.name = name
        def __getitem__(self, x, y, z):
            return f"{x, y, z}"
        def __setitem__(self, value, x, y, z):
            self.set = f"{self.name}[{x, y, z}] = {value}"
        def __delitem__(self, x, y, z):
            self.del_ = f"del {self.name}[{x, y, z}]"
    return C


@pytest.fixture
def c(C_class):
    return C_class("c")


@pytest.fixture
def args():
    return 1,2,3


@pytest.fixture
def value():
    return "foo"


@pytest.fixture
def c_set(c, args, value):
    c[args] = value
    return c


@pytest.fixture
def c_del(c, args):
    del c[args]
    return c


def test_getitem(c, args):
    assert c[args] == f"{(*args,)}"


def test_setitem(c_set, args, value):
    assert c_set.set == f"{c_set.name}[{(*args,)}] = {value}"


def test_delitem(c_del, args):
    assert c_del.del_ == f"del {c_del.name}[{(*args,)}]"



from functools import wraps
from inspect import signature, _empty


def _adapt_get(func):
    @wraps(func)
    def wrapped(self, key):
        return func(self, *key)
    return wrapped


def _adapt_set(func):
    @wraps(func)
    def wrapped(self, key, value):
        func(self, value, *key)
    return wrapped


def _adapt_del(func):
    @wraps(func)
    def wrapped(self, key):
        func(self, *key)
    return wrapped


def _adapt_item_dunders(cls):
    for m_name, adapter, n_params in zip(("getitem", "setitem", "delitem"),
                                         (_adapt_get, _adapt_set, _adapt_del),
                                         (2, 3, 2),
                                         ):
        m_dunder = f"__{m_name}__"
        if method := getattr(cls, m_dunder, None):
            if all(p.default==_empty for p in signature(method).parameters.values()) \
                    and len(signature(method).parameters)==n_params:
                return
            setattr(cls, m_dunder, adapter(method))


class SigDepMeta(type):
    def __init__(cls, name, bases, namespace):
        super().__init__(name, bases, namespace)
        _adapt_item_dunders(cls)


def signature_dependent_semantics(cls):
    _adapt_item_dunders(cls)
    return cls



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['sigdepsem']

package_data = \
{'': ['*']}

extras_require = \
{'test': ['pytest', 'pytest-runner']}

setup(name='sigdepsem',
      version='0.1',
      description='Demonstrate signature dependent semantics for item get, set, del',
      author='Rick L Teachey Jr',
      author_email='ricky@teachey.org',
      url=None,
      packages=packages,
      package_data=package_data,
      extras_require=extras_require,
     )
